@@ -6,11 +6,11 @@ import sys
 from contextlib import contextmanager, suppress
 from copy import copy
 
+from tatsu.collections import BoundedDict
+
 from . import buffering, color, tokenizing
 from .ast import AST
-from .collections import OrderedSet as oset
 from .exceptions import (
-    FailedCut,
     FailedExpectingEndOfText,
     FailedKeywordSemantics,
     FailedLeftRecursion,
@@ -27,11 +27,11 @@ from .infos import (
     Alert,
     MemoKey,
     ParseInfo,
-    ParserConfig,
     ParseState,
     RuleInfo,
     RuleResult,
 )
+from .parserconfig import ParserConfig
 from .tokenizing import Tokenizer
 from .util import (
     info,
@@ -207,7 +207,7 @@ class ParseContext:
         if self.active_config.colorize:
             color.init()
         self._initialize_caches()
-        self._keywords = oset(config.keywords)
+        self._keywords = set(config.keywords)
         self._semantics = config.semantics
         if hasattr(self.semantics, 'set_context'):
             self.semantics.set_context(self)
@@ -239,9 +239,6 @@ class ParseContext:
         try:
             rule = self._find_rule(start)
             return rule()
-        except FailedCut as e:
-            self._set_furthest_exception(e.nested)
-            raise self._furthest_exception from e
         except FailedParse as e:
             self._set_furthest_exception(e)
             raise self._furthest_exception from e
@@ -272,7 +269,7 @@ class ParseContext:
         return self._tokenizer.pos
 
     def _clear_memoization_caches(self):
-        self._memos = {}
+        self._memos = BoundedDict(self.config.memo_cache_size)
         self._results = {}
         self._recursion_depth = 0
 
@@ -433,13 +430,19 @@ class ParseContext:
         # be proven if doing it this way affects linearity. Empirically,
         # it hasn't.
 
-        def prune(cache, cutpos):
-            prune_dict(cache, lambda k, _: k[0] < cutpos)
+        def prune(cache, cut_pos):
+            prune_dict(
+              cache,
+              lambda k, v: k[0] < cut_pos and not isinstance(v, FailedLeftRecursion),
+              )
 
         prune(self._memos, self._pos)
 
     def _memoization(self):
-        return self.memoize_lookaheads or self._lookahead == 0
+        return self.config.memoization and (
+                    self.memoize_lookaheads or
+                    self._lookahead == 0
+                )
 
     def _rulestack(self):
         rulestack = [r.name for r in reversed(self._rule_stack)]
@@ -793,11 +796,11 @@ class ParseContext:
             with self._try():
                 yield
             raise OptionSucceeded()
-        except FailedCut:
-            raise
-        except FailedParse as e:
+        except FailedParse:
             if self._is_cut_set():
-                raise FailedCut(e) from e
+                raise
+            else:
+                pass  # ignore, so next option is tried
         finally:
             self._cut_stack.pop()
 
